{ File   : RESExt.dpr
  Author : Deniz Oezmen
  Created: 2009-08-22
  Changed: 2009-08-22

  Memento Mori RES extractor.
}
program RESExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, Windows, ZLib;

const
  RESHeaderID: WideString = 'Centauri Production Resource File 3.10'#010#010;  

type
  TRESHeader = packed record
    ID        : array[0..40] of WideChar;
    NumEntries: Longint;
    DirOfs    : Longint
  end;

  TRESEntry = record
    Name: WideString;
    BlockData: packed record
      DateStamp : TFileTime;
      CompType,
      Ofs,
      UncompSize,
      DataSize,
      Unknown   : Longint;
    end;
  end;

  TRESEntries = array of TRESEntry;

  TBuffer = array of Byte;

function OpenInFile(var InFile: file; const InName: AnsiString): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; const OutName: AnsiString): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure SetDateStamp(const FileName: AnsiString; DateStamp: TFileTime);
var
  Handle: Longword;
begin
  LocalFileTimeToFileTime(DateStamp, DateStamp);

  Handle := FileOpen(FileName, fmOpenWrite);
  if Handle = INVALID_HANDLE_VALUE then
    Exit;

  SetFileTime(Handle, @DateStamp, @DateStamp, @DateStamp);
  
  FileClose(Handle)
end;

function ReadHeader(var RESFile: file; out RESHeader: TRESHeader): Boolean;
var
  BytesRead: Longint;
begin
  BlockRead(RESFile, RESHeader, SizeOf(RESHeader), BytesRead);
  Result := (BytesRead = SizeOf(RESHeader)) and (RESHeader.ID = RESHeaderID)
end;

function ReadEntries(var RESFile: file; const RESHeader: TRESHeader;
  out RESEntries: TRESEntries): Boolean;
var
  i: Longint;
  w: WideChar;
begin
  {$I-}
  Seek(RESFile, RESHeader.DirOfs);
  {$I+}
  Result := (IOResult = 0);

  if not Result then
  begin
    WriteLn('  directory seek error');
    Exit
  end;

  try
    SetLength(RESEntries, RESHeader.NumEntries);
  except
    on EOutOfMemory do
    begin
      WriteLn('  not enough memory to allocate directory');
      Exit
    end
  end;

  for i := 0 to High(RESEntries) do
  begin
    {$I-}
    repeat
      BlockRead(RESFile, w, SizeOf(w));
      if w <> #000 then
        RESEntries[i].Name := RESEntries[i].Name + w
    until (w = #000) or (Eof(RESFile));

    BlockRead(RESFile, RESEntries[i].BlockData,
      SizeOf(RESEntries[i].BlockData));
    {$I+}
    Result := (IOResult = 0);

    if IOResult <> 0 then
    begin
      WriteLn('directory read error');
      Exit
    end
  end
end;

function DecompressZLib(const InBuffer: TBuffer;
  var OutBuffer: TBuffer): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  DecompressBuf(@InBuffer[0], Length(InBuffer), Length(OutBuffer), TempBuffer,
    DecompSize);

  Result := (DecompSize = Length(OutBuffer));

  if Result then
    Move(TempBuffer^, OutBuffer[0], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

procedure ExtractFiles(var RESFile: file; const RESEntries: TRESEntries);
var
  i,
  AllocSize: Longint;
  InBuf,
  OutBuf   : TBuffer;
  OutFile  : file;
begin
  for i := 0 to High(RESEntries) do
  begin
    WriteLn('  ', RESEntries[i].Name);

    AllocSize := 0;

    try
      // read data
      {$I-}
      Seek(RESFile, RESEntries[i].BlockData.Ofs);
      AllocSize := RESEntries[i].BlockData.DataSize;
      SetLength(InBuf, AllocSize);
      BlockRead(RESFile, InBuf[0], Length(InBuf));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    file read error');
        SetLength(InBuf, 0);
        SetLength(OutBuf, 0);
        Exit
      end;

      // decompress if necessary
      case RESEntries[i].BlockData.CompType of
        0: OutBuf := InBuf;
        2: begin
             AllocSize := RESEntries[i].BlockData.UncompSize;
             SetLength(OutBuf, AllocSize);
             if not DecompressZLib(InBuf, OutBuf) then
             begin
               WriteLn('    decompression error');
               Continue
             end
           end
       else
         WriteLn(Format('    unsupported compression type %d',
           [RESEntries[i].BlockData.CompType]))
       end;

       // output
       ForceDirectories(ExtractFilePath(ExpandFileName(RESEntries[i].Name)));
       if OpenOutFile(OutFile, RESEntries[i].Name) then
       begin
         {$I+}
         BlockWrite(OutFile, OutBuf[0], Length(OutBuf));
         {$I-}
         if IOResult <> 0 then
           WriteLn('    file write error');
         CloseFile(OutFile);
         SetDateStamp(RESEntries[i].Name, RESEntries[i].BlockData.DateStamp)
       end
       else
         WriteLn('    file creation error');
    except
      on EOutOfMemory do
      begin
        WriteLn(Format('    not enough memory to allocate %d bytes',
          [AllocSize]));
        SetLength(InBuf, 0);
        SetLength(OutBuf, 0);
        Exit
      end
    end
  end
end;

procedure AnalyzeFile(var RESFile: file);
var
  RESHeader : TRESHeader;
  RESEntries: TRESEntries;
begin
   if ReadHeader(RESFile, RESHeader) then
   begin
     if ReadEntries(RESFile, RESHeader, RESEntries) then
       ExtractFiles(RESFile, RESEntries);
     SetLength(RESEntries, 0)
   end
   else
     WriteLn('  not a valid RES file')
end;

procedure ScanFiles(FileParam: AnsiString);
var
  Dir,
  InName   : AnsiString;
  InFile   : file;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      InName := Dir + SearchRec.Name;

      WriteLn(InName);

      if OpenInFile(InFile, InName) then
        AnalyzeFile(InFile)
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  SysUtils.FindClose(SearchRec)
end;

begin
  if ParamCount = 1 then
    ScanFiles(ParamStr(1))
  else
    WriteLn('Incorrect number of parameters.')
end.

